"""
Embedding model dimensions and aliases.

This module contains predefined dimensions for various embedding models
and mapping aliases for backward compatibility.
"""

from typing import Optional, List


# Dictionary mapping model names to their embedding dimensions
KNOWN_MODEL_DIMENSIONS = {
    # TiDB Cloud Free models
    "tidbcloud_free/amazon/titan-embed-text-v2": 1024,
    "tidbcloud_free/cohere/embed-english-v3": 1024,
    "tidbcloud_free/cohere/embed-multilingual-v3": 1024,
    # OpenAI models
    "openai/text-embedding-3-small": 1536,
    "openai/text-embedding-3-large": 3072,
    "openai/text-embedding-ada-002": 1536,
    # Cohere models
    "cohere/embed-english-v3.0": 1024,
    "cohere/embed-multilingual-v3.0": 1024,
    # Jina AI models
    "jina_ai/jina-embeddings-v4": 2048,
    "jina_ai/jina-embeddings-v3": 1024,
    "jina_ai/jina-clip-v2": 1024,
    # TODO: remove these after jina_ai is released on prod.
    "jina/jina-embeddings-v4": 2048,
    "jina/jina-embeddings-v3": 1024,
    "jina/jina-clip-v2": 1024,
}

# Mapping of model aliases to their full names for backward compatibility
MODEL_ALIASES = {
    "text-embedding-3-small": "openai/text-embedding-3-small",
    "text-embedding-3-large": "openai/text-embedding-3-large",
    "text-embedding-ada-002": "openai/text-embedding-ada-002",
}


def get_model_dimensions(model_name: str) -> Optional[int]:
    normalized_name = MODEL_ALIASES.get(model_name, model_name)
    return KNOWN_MODEL_DIMENSIONS.get(normalized_name)


def is_known_model(model_name: str) -> bool:
    normalized_name = MODEL_ALIASES.get(model_name, model_name)
    return normalized_name in KNOWN_MODEL_DIMENSIONS


def list_known_models() -> List[str]:
    return list(KNOWN_MODEL_DIMENSIONS.keys()) + list(MODEL_ALIASES.keys())


def register_model_dimension(model_name: str, dimensions: int):
    KNOWN_MODEL_DIMENSIONS[model_name] = dimensions


def register_model_alias(alias: str, canonical_name: str):
    MODEL_ALIASES[alias] = canonical_name



from contextlib import contextmanager
from contextvars import ContextVar
from typing import List, Literal, Optional, Type, Generator

from pydantic import PrivateAttr
import sqlalchemy
from sqlalchemy import (
    Executable,
    Inspector,
    SelectBase,
    text,
    Result,
)
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.orm import Session, DeclarativeMeta

from pytidb.orm.variables import EMBED_PROVIDER_API_KEY_VARS
from pytidb.base import Base, default_registry
from pytidb.databases import create_database, database_exists
from pytidb.schema import TableModel
from pytidb.table import Table
from pytidb.utils import (
    TIDB_SERVERLESS_HOST_PATTERN,
    build_tidb_connection_url,
    create_engine_without_db,
)
from pytidb.logger import logger
from pytidb.result import SQLExecuteResult, SQLQueryResult


SESSION = ContextVar[Session | None]("session", default=None)


class TiDBClient:
    _db_engine: Engine = PrivateAttr()
    _inspector: Inspector = PrivateAttr()

    # Necessary connection parameters for use_database functionality
    _reconnect_params: dict = PrivateAttr()

    def __init__(self, db_engine: Engine, reconnect_params: dict):
        self._db_engine = db_engine
        self._inspector = sqlalchemy.inspect(self._db_engine)
        self._identifier_preparer = self._db_engine.dialect.identifier_preparer
        self._reconnect_params = reconnect_params

    # TODO: Better typing for kwargs, including what's supported by pymysql and SQLAlchemy.
    @classmethod
    def connect(
        cls,
        url: Optional[str] = None,
        *,
        host: Optional[str] = "localhost",
        port: Optional[int] = 4000,
        username: Optional[str] = "root",
        password: Optional[str] = "",
        database: Optional[str] = "test",
        enable_ssl: Optional[bool] = None,
        ensure_db: Optional[bool] = False,
        debug: Optional[bool] = None,
        **kwargs,
    ) -> "TiDBClient":
        if url is None:
            url = build_tidb_connection_url(
                host=host,
                port=port,
                username=username,
                password=password,
                database=database,
                enable_ssl=enable_ssl,
            )
            # TODO: When URL is passed in directly, it should be validated.

        if ensure_db:
            try:
                temp_engine = create_engine_without_db(url, echo=debug, **kwargs)
                if not database_exists(temp_engine, database):
                    create_database(temp_engine, database)
            except Exception as e:
                logger.error("Failed to ensure database exists: %s", str(e))
                raise

        if host and TIDB_SERVERLESS_HOST_PATTERN.match(host):
            kwargs.setdefault("pool_recycle", 300)
            kwargs.setdefault("pool_pre_ping", True)
            kwargs.setdefault("pool_timeout", 10)

        db_engine = create_engine(url, echo=debug, **kwargs)
        reconnect_params = {
            # host, port, etc is not needed because they will be built into the URL.
            # url is also not needed because it is already in `db_engine`.
            "ensure_db": ensure_db,
            "debug": debug,
            **kwargs,
        }

        return cls(db_engine, reconnect_params)

    def disconnect(self) -> None:
        self._db_engine.dispose()

    @property
    def db_engine(self) -> Engine:
        return self._db_engine

    # Database Management API

    def create_database(
        self, name: str, if_exists: Optional[Literal["raise", "skip"]] = "raise"
    ):
        return create_database(self._db_engine, name, if_exists=if_exists)

    def drop_database(self, name: str):
        db_name = self._identifier_preparer.quote(name)
        with self._db_engine.connect() as conn:
            stmt = text(f"DROP DATABASE IF EXISTS {db_name};")
            return conn.execute(stmt)

    def list_databases(self) -> List[str]:
        stmt = text("SHOW DATABASES;")
        with self._db_engine.connect() as conn:
            result = conn.execute(stmt)
            return [row[0] for row in result]

    def has_database(self, name: str) -> bool:
        return database_exists(self._db_engine, name)

    def current_database(self) -> Optional[str]:
        """Get the current database name.

        Returns:
            The name of the current database, or None if no database is selected.
        """
        stmt = text("SELECT DATABASE();")
        with self._db_engine.connect() as conn:
            result = conn.execute(stmt)
            return result.scalar()

    def use_database(self, database: str, *, ensure_db: Optional[bool] = False) -> None:
        """Switch to a different database.

        This method provides the same experience as if the database was specified
        when calling connect(). It creates a new engine with the specified database
        to ensure the database context persists across connection drops and reconnects.

        Warning: Existing sessions will be destroyed.

        Args:
            database: The name of the database to switch to.
            ensure_db: If True, create the database if it doesn't exist.

        Raises:
            Exception: If the database doesn't exist and ensure_db is False.
        """

        # TODO: Find some way to remove the restriction that old sessions will be destroyed.
        # Generally we should allow using `use_database` in a chainable way where it opens a new
        # "database session", for example, two lines below should be allowed to work simultaneously:
        # client.use_database(a).create_table()
        # client.use_database(b).create_table()

        if ensure_db and not self.has_database(database):
            self.create_database(database, if_exists="skip")
        elif not ensure_db and not self.has_database(database):
            raise ValueError(f"Database '{database}' does not exist")

        new_url = self._db_engine.url.set(database=database)

        # Attempt to create the new client first, only dispose and update attributes if successful
        new_client = TiDBClient.connect(
            url=new_url.render_as_string(hide_password=False),
            **self._reconnect_params,
        )

        # Now that new_client is successfully created, dispose the old engine and update all attributes
        self._db_engine.dispose()
        self._db_engine = new_client._db_engine
        self._inspector = new_client._inspector
        self._identifier_preparer = new_client._db_engine.dialect.identifier_preparer
        self._reconnect_params = new_client._reconnect_params

    # Table Management API

    def create_table(
        self,
        *,
        schema: Optional[Type[TableModel]] = None,
        if_exists: Optional[Literal["raise", "overwrite", "skip"]] = "raise",
    ) -> Table:
        if if_exists == "raise":
            table = Table(schema=schema, client=self, if_exists="raise")
        elif if_exists == "overwrite":
            self.drop_table(schema.__tablename__, if_not_exists="skip")
            table = Table(schema=schema, client=self)
        elif if_exists == "skip":
            table = Table(schema=schema, client=self, if_exists="skip")
        else:
            raise ValueError(f"Invalid if_exists value: {if_exists}")
        return table

    def _get_table_model(self, table_name: str) -> Optional[Type[DeclarativeMeta]]:
        for m in default_registry.mappers:
            if m.persist_selectable.name == table_name:
                return m.class_
        return None

    def open_table(self, table_name: str) -> Optional[Table]:
        # If the table in the mapper registry.
        table_model = self._get_table_model(table_name)
        if table_model is not None:
            table = Table(schema=table_model, client=self, if_exists="skip")
            return table

        return None

    def list_tables(self) -> List[str]:
        stmt = text("SHOW TABLES;")
        with self._db_engine.connect() as conn:
            result = conn.execute(stmt)
            return [row[0] for row in result]

    def has_table(self, table_name: str) -> bool:
        return self._inspector.has_table(table_name)

    def drop_table(
        self,
        table_name: str,
        if_not_exists: Optional[Literal["raise", "skip"]] = "raise",
    ):
        if if_not_exists not in ["raise", "skip"]:
            raise ValueError(f"Invalid if_not_exists value: {if_not_exists}")

        table = sqlalchemy.Table(
            table_name, Base.metadata, autoload_with=self._db_engine
        )
        return table.drop(self._db_engine, checkfirst=(if_not_exists == "skip"))

    # Raw SQL API

    def execute(
        self,
        sql: str | Executable,
        params: Optional[dict] = None,
        raise_error: Optional[bool] = False,
    ) -> SQLExecuteResult:
        try:
            with self.session() as session:
                if isinstance(sql, str):
                    stmt = text(sql)
                else:
                    stmt = sql
                result: Result = session.execute(stmt, params or {})
                return SQLExecuteResult(rowcount=result.rowcount, success=True)
        except Exception as e:
            if raise_error:
                raise e
            logger.error(f"Failed to execute SQL: {str(e)}")
            return SQLExecuteResult(rowcount=0, success=False, message=str(e))

    def query(
        self,
        sql: str | SelectBase,
        params: Optional[dict] = None,
    ) -> SQLQueryResult:
        with self.session() as session:
            if isinstance(sql, str):
                stmt = text(sql)
            else:
                stmt = sql
            result = session.execute(stmt, params)
            return SQLQueryResult(result)

    def configure_embedding_provider(
        self, provider: str, api_key: str
    ) -> SQLExecuteResult:
        if provider not in EMBED_PROVIDER_API_KEY_VARS.keys():
            raise ValueError(
                f"Unsupported configure api key for embedding provider: {provider}"
            )

        var_name = EMBED_PROVIDER_API_KEY_VARS[provider]
        escape_var_name = self._identifier_preparer.quote(var_name)
        return self.execute(text(f"SET @@GLOBAL.{escape_var_name} = '{api_key}';"))

    # Session API

    @contextmanager
    def session(
        self, *, provided_session: Optional[Session] = None, **kwargs
    ) -> Generator[Session, None, None]:
        if provided_session is not None:
            session = provided_session
            is_local_session = False
        elif SESSION.get() is not None:
            session = SESSION.get()
            is_local_session = False
        else:
            # Since both the TiDB Client and Table API begin a Session within the method, the Session ends when
            # the method returns. The error: "Parent instance <x> is not bound to a Session;" will show when accessing
            # the returned object. To prevent it, we set the expire_on_commit parameter to False by default.
            # Details: https://sqlalche.me/e/20/bhk3
            kwargs.setdefault("expire_on_commit", False)
            session = Session(self._db_engine, **kwargs)
            SESSION.set(session)
            is_local_session = True

        try:
            yield session
            if is_local_session:
                session.commit()
        except Exception as e:
            session.rollback()
            raise e
        finally:
            if is_local_session:
                session.close()
                SESSION.set(None)



import math
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple

from sqlalchemy import result_tuple
from sqlalchemy.engine import Row


from pytidb.utils import RowKeyType
from pytidb.schema import DistanceMetric


FusionFunction = Callable[[Any, Any, Row, Row, Optional[RowKeyType]], Any]


def merge_result_rows(
    rows_a: List[Row],
    rows_b: List[Row],
    get_row_key: Callable[[Row], RowKeyType],
    fusion_strategies: Optional[Dict[str, FusionFunction]] = None,
) -> Tuple[List[str], List[Row]]:
    """Merge two lists of result rows based on row_id.

    Args:
        rows_a: First list of result rows
        rows_b: Second list of result rows
        get_row_key: Function to get the key (primary key or _tidb_rowid) from a row
        fusion_strategies: Optional dictionary mapping field names to custom fusion functions.
                   Each fusion function takes (value_a, value_b, row_a, row_b, key) as arguments
                   and returns the fused value. If both rows_a and rows_b are empty,
                   fusion_strategies will not be used.

    Returns:
        Tuple containing:
        - List of all field names from both input rows
        - List of merged result rows
    """
    if not rows_a and not rows_b:
        return [], []

    # Get all column names
    fields_a = list(rows_a[0]._fields) if len(rows_a) > 0 else []
    fields_b = list(rows_b[0]._fields) if len(rows_b) > 0 else []
    all_fields = list(dict.fromkeys(fields_a + fields_b).keys())

    # Create a mapping of keys to rows from both lists
    rows_by_key_a = {get_row_key(row): row for row in rows_a}
    rows_by_key_b = {get_row_key(row): row for row in rows_b}

    # Get all unique keys.
    all_keys = set(rows_by_key_a.keys()) | set(rows_by_key_b.keys())

    # Merge results
    merged_rows = []
    for key in all_keys:
        row_data = []
        row_a = rows_by_key_a.get(key)
        row_b = rows_by_key_b.get(key)

        for field in all_fields:
            value_a = getattr(row_a, field) if row_a and field in fields_a else None
            value_b = getattr(row_b, field) if row_b and field in fields_b else None

            # Use custom merge strategy if provided
            if fusion_strategies and field in fusion_strategies:
                value = fusion_strategies[field](value_a, value_b, row_a, row_b, key)
            else:
                # Default strategy: use value_a if not None, otherwise use value_b
                value = value_a if value_a is not None else value_b

            row_data.append(value)

        # Create new Row object using result_tuple
        row_factory = result_tuple(all_fields)
        merged_row = row_factory(row_data)
        merged_rows.append(merged_row)

    return all_fields, merged_rows


def fusion_result_rows_by_rrf(
    rows_a: List[Row],
    rows_b: List[Row],
    get_row_key: Callable[[Row], RowKeyType],
    k: Optional[int] = 60,
) -> Tuple[List[str], List[Row]]:
    """
    Fusion the search results by RRF (Reciprocal Rank Fusion).

    Args:
        rows_a: First list of result rows
        rows_b: Second list of result rows
        get_row_key: Function to get the key (primary key or _tidb_rowid) from a row
        k: The constant used in RRF formula. Must be a positive number. Default is 60.

    Returns:
        Tuple containing:
        - List of field names
        - List of fused result rows sorted by RRF score
    """
    if not rows_a and not rows_b:
        return [], []

    if k <= 0:
        raise ValueError("k must be a positive number")

    # Calculate RRF scores for each result in both lists
    rrf_scores = {}

    # Process first list
    for i, row in enumerate(rows_a):
        rank = i + 1
        key = get_row_key(row)
        rrf_scores[key] = 1.0 / (k + rank)

    # Process second list and add scores
    for i, row in enumerate(rows_b):
        rank = i + 1
        key = get_row_key(row)
        if key in rrf_scores:
            rrf_scores[key] += 1.0 / (k + rank)
        else:
            rrf_scores[key] = 1.0 / (k + rank)

    # Merge rows.
    fusion_strategies = {
        "_score": lambda a, b, row_a, row_b, key: rrf_scores[key],
    }
    all_fields, merged_rows = merge_result_rows(
        rows_a, rows_b, get_row_key, fusion_strategies
    )

    # Sort rows by RRF score.
    sorted_rows = sorted(
        merged_rows, key=lambda row: row._mapping["_score"] or 0, reverse=True
    )

    return all_fields, sorted_rows


def fusion_result_rows_by_weighted(
    vs_rows: List[Row],
    fts_rows: List[Row],
    get_row_key: Callable[[Row], RowKeyType],
    vs_metric: DistanceMetric,
    vs_weight: float = 0.5,
    fts_weight: float = 0.5,
) -> Tuple[List[str], List[Row]]:
    """
    Fusion the search results by weights.

    Args:
        vs_rows: Vector search list of result rows
        fts_rows: Full text search list of result rows
        get_row_key: Function to get the key (primary key or _tidb_rowid) from a row
        vs_metric: The metric type of vector search.
        vs_weight: The weight of the vector search results. Must be a number between 0 and 1. Default is 0.5.
        fts_weight: The weight of the full text search results. Must be a number between 0 and 1. Default is 0.5.
    Returns:
        Tuple containing:
        - List of field names
        - List of fused result rows sorted by RRF score
    """
    if not vs_rows and not fts_rows:
        return [], []

    if vs_weight <= 0 or vs_weight >= 1 or fts_weight <= 0 or fts_weight >= 1:
        raise ValueError("weight must be a number between 0 and 1")
    if vs_weight + fts_weight == 0:
        raise ValueError("At least one weight must be greater than 0")

    # Calculate weighted scores for each result in both lists
    weighted_scores = {}

    # Process first list
    for i, row in enumerate(vs_rows):
        vs_distance = row._mapping["_distance"]

        if vs_metric == DistanceMetric.COSINE:
            normalized_vs_distance = _normalize_score(vs_distance, "cosine")
        elif vs_metric == DistanceMetric.L2:
            normalized_vs_distance = _normalize_score(vs_distance, "l2")
        else:
            raise ValueError("Invalid distance metric")

        key = get_row_key(row)
        weighted_scores[key] = normalized_vs_distance * vs_weight

    # Process second list and add scores
    for i, row in enumerate(fts_rows):
        match_score = row._mapping["_match_score"]
        normalized_match_score = _normalize_score(match_score, "bm25")
        key = get_row_key(row)
        if key in weighted_scores:
            weighted_scores[key] += normalized_match_score * fts_weight
        else:
            weighted_scores[key] = normalized_match_score * fts_weight

    # Merge rows.
    fusion_strategies = {
        "_score": lambda a, b, row_a, row_b, key: weighted_scores[key],
    }
    all_fields, merged_rows = merge_result_rows(
        vs_rows, fts_rows, get_row_key, fusion_strategies
    )

    # Sort rows by weighted score.
    sorted_rows = sorted(
        merged_rows, key=lambda row: row._mapping["_score"] or 0, reverse=True
    )

    return all_fields, sorted_rows


def _normalize_score(
    score: float, metric_type: Literal["l2", "cosine", "bm25"]
) -> float:
    """Normalize the score to [0, 1] range and ensure that the higher normalized values
    correspond to greater similarity.

    Args:
        score: The score to normalize.
        metric_type: The metric type of score.

    Returns:
        The normalized score between 0 and 1.
    """

    if metric_type == "cosine":
        # Cosine distance range is [0, 2] and smaller values indicate higher
        # similarity (need inverted).
        return 1.0 - (score / 2.0)
    elif metric_type == "l2":
        # L2 distance range is [0, +infinity) and smaller values indicate higher
        # similarity (need inverted).
        return 1.0 - (2.0 * math.atan(score) / math.pi)
    elif metric_type == "bm25":
        # BM25 score range is [0, +infinity) and larger values indicate higher
        # match degree (no need to invert).
        return 2.0 * math.atan(score) / math.pi
    else:
        raise ValueError(f"Invalid metric type for score normalization: {metric_type}")



from typing import Any, Literal, Optional, TYPE_CHECKING, List, TypedDict, Union
import json

from pydantic import BaseModel
from sqlalchemy import Column, Computed, Index
from sqlmodel import SQLModel, Field, Relationship
from sqlmodel.main import FieldInfo, RelationshipInfo, SQLModelMetaclass

from pytidb.sql import func
from pytidb.orm.types import TEXT, VECTOR
from pytidb.orm.indexes import VectorIndexAlgorithm
from pytidb.orm.distance_metric import DistanceMetric, validate_distance_metric


if TYPE_CHECKING:
    from pytidb.embeddings.base import BaseEmbeddingFunction, EmbeddingSourceType


VectorDataType = List[float]

IndexType = Literal["vector", "fulltext", "scalar"]


class QueryBundle(TypedDict):
    query: Optional[Any]
    query_vector: Optional[VectorDataType]


class TableModelMeta(SQLModelMetaclass):
    def __new__(mcs, name, bases, namespace, **kwargs):
        if name != "TableModel":
            kwargs.setdefault("table", True)
        return super().__new__(mcs, name, bases, namespace, **kwargs)


class TableModel(SQLModel, metaclass=TableModelMeta):
    pass


Field = Field
Relationship = Relationship
Column = Column
Index = Index
FieldInfo = FieldInfo
RelationshipInfo = RelationshipInfo


def VectorField(
    dimensions: int,
    source_field: Optional[str] = None,
    embed_fn: Optional["BaseEmbeddingFunction"] = None,
    source_type: "EmbeddingSourceType" = "text",
    index: Optional[bool] = None,
    distance_metric: Optional[Union[DistanceMetric, str]] = DistanceMetric.COSINE,
    algorithm: Optional[VectorIndexAlgorithm] = "HNSW",
    **kwargs,
):
    # Notice: Currently, only L2 and COSINE distance metrics support indexing.
    distance_metric = validate_distance_metric(distance_metric)
    if index is None:
        if distance_metric in [DistanceMetric.L2, DistanceMetric.COSINE]:
            index = True
        else:
            index = False

    use_server = embed_fn.use_server if embed_fn else False
    if use_server:
        model_name = embed_fn.model_name
        embed_params = {
            **(embed_fn.server_embed_params or {}),
            **kwargs.get("server_embed_params", {}),
        }

        source_column = Column(source_field)
        if embed_params:
            embed_params_str = json.dumps(embed_params)
            embed_sql_function = func.EMBED_TEXT(
                model_name, source_column, embed_params_str
            )
        else:
            embed_sql_function = func.EMBED_TEXT(model_name, source_column)

        default_sa_column = Column(
            VECTOR(dimensions),
            Computed(embed_sql_function, persisted=True),
        )
    else:
        default_sa_column = Column(VECTOR(dimensions))

    sa_column = kwargs.pop("sa_column", default_sa_column)

    return Field(
        sa_column=sa_column,
        schema_extra={
            "field_type": "vector",
            "dimensions": dimensions,
            # Auto embedding related.
            "embed_fn": embed_fn,
            "source_field": source_field,
            "source_type": source_type,
            "use_server": use_server,
            # Vector index related.
            "skip_index": not index,
            "distance_metric": distance_metric,
            "algorithm": algorithm,
        },
        **kwargs,
    )


def FullTextField(
    index: Optional[bool] = True,
    fts_parser: Optional[str] = "MULTILINGUAL",
    **kwargs,
):
    sa_column = kwargs.pop("sa_column", Column(TEXT))
    return Field(
        sa_column=sa_column,
        schema_extra={
            "field_type": "text",
            # Fulltext index related.
            "skip_index": not index,
            "fts_parser": fts_parser,
        },
        **kwargs,
    )


class ColumnInfo(BaseModel):
    column_name: str
    column_type: str



from abc import ABC, abstractmethod
from typing import Optional, Any, Literal

from pydantic import BaseModel
from sqlmodel import Field


EmbeddingSourceType = Literal["text", "image"]


class BaseEmbeddingFunction(BaseModel, ABC):
    provider: str = Field("openai", description="The name of the embedding provider")
    model_name: str = Field(
        None, description="The name of embedding model used for embedding"
    )
    dimensions: Optional[int] = Field(
        None, description="Dimensions of the vector generated by the embedding model"
    )
    use_server: bool = Field(
        False,
        description=(
            "Whether to enable server-side embedding (auto embedding in the database). "
            "If True, the embedding for the source data will be computed in the database. "
            "Default is False."
        ),
    )
    server_embed_params: Optional[dict[str, Any]] = Field(
        None, description="The parameters for server-side embedding. "
    )

    def __init__(self, /, **data: Any):
        super().__init__(**data)

    def VectorField(
        self,
        source_field: Optional[str] = None,
        source_type: Optional[EmbeddingSourceType] = "text",
        **kwargs,
    ):
        """
        Create a VectorField with auto embedding configuration.

        Args:
            source_field: The name of the source field to embed
            source_type: The type of source data ("text" or "image")
            **kwargs: Additional keyword arguments for VectorField

        Returns:
            VectorField configured with this embedding function
        """
        # Import here to avoid circular import
        from pytidb.schema import VectorField

        return VectorField(
            embed_fn=self,
            dimensions=self.dimensions,
            source_field=source_field,
            source_type=source_type,
            **kwargs,
        )

    @abstractmethod
    def get_query_embedding(
        self, query: Any, source_type: Optional[EmbeddingSourceType] = "text", **kwargs
    ) -> list[float]:
        """
        Get embedding for a query (text or image).

        Args:
            query: Query text string or PIL Image object
            source_type: The type of source data ("text" or "image")

        Returns:
            List of float values representing the embedding
        """
        raise NotImplementedError()

    @abstractmethod
    def get_source_embedding(
        self, source: Any, source_type: Optional[EmbeddingSourceType] = "text", **kwargs
    ) -> list[float]:
        """
        Get embedding for a source field value (typically text).

        Args:
            source: Source field value (text)
            source_type: The type of source data ("text" or "image")

        Returns:
            List of float values representing the embedding
        """
        raise NotImplementedError()

    @abstractmethod
    def get_source_embeddings(
        self,
        sources: list[Any],
        source_type: Optional[EmbeddingSourceType] = "text",
        **kwargs,
    ) -> list[list[float]]:
        """
        Get embeddings for multiple source field values.

        Args:
            sources: List of source field values
            source_type: The type of source data ("text" or "image")

        Returns:
            List of embeddings, where each embedding is a list of float values
        """
        raise NotImplementedError()



from sqlalchemy.sql.ddl import SchemaGenerator, CreateIndex
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import elements, operators, functions


@compiles(CreateIndex, "mysql")
def compile_create_index(create, compiler, **kw):
    # Copy from sqlalchemy.dialects.mysql.base.MySQLCompiler::visit_create_index
    index = create.element
    compiler._verify_index_table(index)
    preparer = compiler.preparer
    table = preparer.format_table(index.table)

    columns = [
        compiler.sql_compiler.process(
            (
                elements.Grouping(expr)
                if (
                    isinstance(expr, elements.BinaryExpression)
                    or (
                        isinstance(expr, elements.UnaryExpression)
                        and expr.modifier not in (operators.desc_op, operators.asc_op)
                    )
                    or isinstance(expr, functions.FunctionElement)
                )
                else expr
            ),
            include_table=False,
            literal_binds=True,
        )
        for expr in index.expressions
    ]

    name = compiler._prepared_index_name(index)

    text = "CREATE "
    if index.unique:
        text += "UNIQUE "

    index_prefix = index.kwargs.get("%s_prefix" % compiler.dialect.name, None)
    if index_prefix:
        text += index_prefix + " "

    text += "INDEX "
    if create.if_not_exists:
        text += "IF NOT EXISTS "
    text += "%s ON %s " % (name, table)

    length = index.dialect_options[compiler.dialect.name]["length"]
    if length is not None:
        if isinstance(length, dict):
            # length value can be a (column_name --> integer value)
            # mapping specifying the prefix length for each column of the
            # index
            columns = ", ".join(
                (
                    "%s(%d)" % (expr, length[col.name])
                    if col.name in length
                    else (
                        "%s(%d)" % (expr, length[expr])
                        if expr in length
                        else "%s" % expr
                    )
                )
                for col, expr in zip(index.expressions, columns)
            )
        else:
            # or can be an integer value specifying the same
            # prefix length for all columns of the index
            columns = ", ".join("%s(%d)" % (col, length) for col in columns)
    else:
        columns = ", ".join(columns)

    text += "(%s)" % columns

    parser = index.dialect_options[compiler.dialect.name]["with_parser"]
    if parser is not None:
        text += " WITH PARSER %s" % (parser,)

    using = index.dialect_options[compiler.dialect.name]["using"]
    if using is not None:
        text += " USING %s" % (using)

    if hasattr(index, "ensure_columnar_replica") and index.ensure_columnar_replica:
        text += " ADD_COLUMNAR_REPLICA_ON_DEMAND"

    return text


class CreateVectorIndex(CreateIndex):
    def __init__(self, element, if_not_exists=False):
        super().__init__(element, if_not_exists=if_not_exists)


class TiDBSchemaGenerator(SchemaGenerator):
    def visit_vector_index(self, index, create_ok=False):
        if not create_ok and not self._can_create_index(index):
            return
        with self.with_ddl_events(index):
            CreateVectorIndex(index)._invoke_with(self.connection)


@compiles(CreateVectorIndex)
def compile_create_vector_index(element, compiler, **kw):
    index = element.index
    table_name = index.table.name
    column_name = index.columns[0].name
    distance_metric = index.distance_metric
    algorithm = index.algorithm
    distance_metric_str = f"({distance_metric.to_sql_func()}({column_name}))"

    sql = f"CREATE VECTOR INDEX {index.name} ON {table_name} ({distance_metric_str}) USING {algorithm}"
    if index.ensure_columnar_replica:
        sql += " ADD_COLUMNAR_REPLICA_ON_DEMAND"
    return sql



from sqlalchemy.sql.functions import GenericFunction


class fts_match_word(GenericFunction):
    type = None
    name = "FTS_MATCH_WORD"
    inherit_cache = True


class vec_l1_distance(GenericFunction):
    type = None
    name = "VEC_L1_DISTANCE"
    inherit_cache = True


class vec_l2_distance(GenericFunction):
    type = None
    name = "VEC_L2_DISTANCE"
    inherit_cache = True


class vec_cosine_distance(GenericFunction):
    type = None
    name = "VEC_COSINE_DISTANCE"
    inherit_cache = True


class vec_negative_inner_product(GenericFunction):
    type = None
    name = "VEC_NEGATIVE_INNER_PRODUCT"
    inherit_cache = True


class vec_embed_l1_distance(GenericFunction):
    type = None
    name = "VEC_EMBED_L1_DISTANCE"
    inherit_cache = True


class vec_embed_l2_distance(GenericFunction):
    type = None
    name = "VEC_EMBED_L2_DISTANCE"
    inherit_cache = True


class vec_embed_cosine_distance(GenericFunction):
    type = None
    name = "VEC_EMBED_COSINE_DISTANCE"
    inherit_cache = True



from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Optional, Union

from pydantic import Field
from pytidb.embeddings.base import BaseEmbeddingFunction, EmbeddingSourceType
from pytidb.embeddings.dimensions import get_model_dimensions
from pytidb.embeddings.utils import (
    encode_local_file_to_base64,
    encode_pil_image_to_base64,
    parse_url_safely,
)
import urllib.request


if TYPE_CHECKING:
    from PIL.Image import Image


# Map of model name -> maximum allowed base64 length (characters)
_MAX_B64_LENGTH_PER_MODEL = {
    # Despite the document says the input image size is up to 25MB,
    # according to the https://docs.aws.amazon.com/bedrock/latest/userguide/titan-multiemb-models.html,
    # the actual limit is 100k for base64 encoded string.
    "bedrock/amazon.titan-embed-image-v1": 100000,
}


PROVIDER_DEFAULT_EMBED_PARAMS = {
    "jina_ai": {
        "task": "retrieval.passage",
        "task@search": "retrieval.query",
    },
}


EmbeddingInput = Union[str, Path, "Image"]


class BuiltInEmbeddingFunction(BaseEmbeddingFunction):
    api_key: Optional[str] = Field(None, description="The API key for authentication.")
    api_base: Optional[str] = Field(
        None, description="The base URL of the model provider."
    )
    timeout: Optional[int] = Field(
        None, description="The timeout value for the API call."
    )
    caching: bool = Field(
        True, description="Whether to cache the embeddings, default True."
    )
    multimodal: bool = Field(
        False,
        description=(
            "Indicates whether the embedding function supports multi-modal input (e.g., both text and image). "
            "Currently, multi-modal embedding is only supported for client-side embedding. "
            "- For multimodal=True, enable client-side embedding by default."
            "- For multimodal=False, disable server-side embedding by default."
        ),
    )

    def __init__(
        self,
        model_name: str,
        dimensions: Optional[int] = None,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        timeout: Optional[int] = None,
        caching: bool = True,
        use_server: Optional[bool] = None,
        server_embed_params: Optional[dict[str, Any]] = None,
        multimodal: bool = False,
        **kwargs,
    ):
        if use_server is None:
            # If multimodal is True, use server-side embedding by default.
            use_server = not multimodal

        if dimensions is None:
            dimensions = get_model_dimensions(model_name)

        provider = model_name.split("/")[0] if "/" in model_name else "openai"
        server_embed_params = (
            server_embed_params
            if server_embed_params is not None
            else PROVIDER_DEFAULT_EMBED_PARAMS.get(provider)
        )

        super().__init__(
            model_name=model_name,
            provider=provider,
            dimensions=dimensions,
            api_key=api_key,
            api_base=api_base,
            timeout=timeout,
            caching=caching,
            use_server=use_server,
            server_embed_params=server_embed_params,
            multimodal=multimodal,
            **kwargs,
        )
        if self.dimensions is None:
            if use_server:
                raise ValueError(
                    f"Missing dimensions for model {self.model_name}. Please specify dimensions."
                )
            else:
                # For client-side embedding, try to infer dimensions with a test embedding.
                embedding = self.get_query_embedding("test", "text")
                if embedding is None:
                    raise ValueError(
                        f"Cannot infer dimensions for model {self.model_name}. Please specify dimensions."
                    )
                self.dimensions = len(embedding)

    def _process_input(
        self, input: EmbeddingInput, source_type: Optional[EmbeddingSourceType] = "text"
    ) -> Union[str, dict]:
        if source_type == "text":
            return input
        elif source_type == "image":
            return self._process_image_input(input)
        else:
            raise ValueError(f"Invalid source type: {source_type}")

    def _process_image_input(self, input: EmbeddingInput) -> Union[str, dict]:
        try:
            from PIL.Image import Image
        except ImportError:
            raise ImportError(
                "PIL (Pillow) is required for image processing. Install it with: pip install Pillow"
            )

        if isinstance(input, Path):
            input = input.resolve().as_uri()

        if isinstance(input, str):
            is_valid, image_url = parse_url_safely(input)
            if is_valid:
                if image_url.scheme == "file":
                    file_path = urllib.request.url2pathname(image_url.path)
                    max_len = _MAX_B64_LENGTH_PER_MODEL.get(self.model_name)
                    base64_str = encode_local_file_to_base64(
                        file_path, max_base64_length=max_len
                    )
                    # For bedrock models, prepend data URL prefix and return string
                    if self.model_name.startswith("bedrock/"):
                        return f"data:image/jpeg;base64,{base64_str}"
                    return {"image": base64_str}
                elif image_url.scheme == "http" or image_url.scheme == "https":
                    # For bedrock models, Bedrock API expects base64 not URL; fall back to query string.
                    if self.model_name.startswith("bedrock/"):
                        return image_url.geturl()
                    return {"image": image_url.geturl()}
                else:
                    raise ValueError(
                        f"invalid url schema for image source: {image_url.scheme}"
                    )
            else:
                return input
        elif isinstance(input, Image):
            max_len = _MAX_B64_LENGTH_PER_MODEL.get(self.model_name)
            base64_str = encode_pil_image_to_base64(input, max_base64_length=max_len)
            if self.model_name.startswith("bedrock/"):
                return f"data:image/jpeg;base64,{base64_str}"
            return {"image": base64_str}
        else:
            raise ValueError(
                "invalid input for image vector search, current supported input types: "
                "url string, Path object, PIL.Image object"
            )

    def get_query_embedding(
        self,
        query: EmbeddingInput,
        source_type: Optional[EmbeddingSourceType] = "text",
        **kwargs,
    ) -> list[float]:
        """
        Get embedding for a query (text or image).

        Args:
            query: Query text string or PIL Image object
            source_type: The type of source data ("text" or "image")

        Returns:
            List of float values representing the embedding
        """
        embedding_input = self._process_input(query, source_type)
        embeddings = self._call_embeddings_api([embedding_input], **kwargs)
        return embeddings[0]

    def get_source_embedding(
        self,
        source: EmbeddingInput,
        source_type: Optional[EmbeddingSourceType] = "text",
        **kwargs,
    ) -> list[float]:
        """
        Get embedding for a source field value (typically text).

        Args:
            source: Source field value (text)
            source_type: The type of source data ("text" or "image")

        Returns:
            List of float values representing the embedding
        """
        embedding_input = self._process_input(source, source_type)
        embeddings = self._call_embeddings_api([embedding_input], **kwargs)
        return embeddings[0]

    def get_source_embeddings(
        self,
        sources: List[EmbeddingInput],
        source_type: Optional[EmbeddingSourceType] = "text",
        **kwargs,
    ) -> list[list[float]]:
        """
        Get embeddings for multiple source field values.

        Args:
            sources: List of source field values
            source_type: The type of source data ("text" or "image")

        Returns:
            List of embeddings, where each embedding is a list of float values
        """
        embedding_inputs = [
            self._process_input(source, source_type) for source in sources
        ]
        embeddings = self._call_embeddings_api(embedding_inputs, **kwargs)
        return embeddings

    def __call__(
        self,
        input: EmbeddingInput,
        source_type: Optional[EmbeddingSourceType] = "text",
        **kwargs,
    ) -> list[float]:
        return self._call_embeddings_api([input], **kwargs)

    def _call_embeddings_api(
        self, input: List[EmbeddingInput], **kwargs
    ) -> List[List[float]]:
        """
        Retrieve embeddings for a given list of input strings using the specified model.

        Args:
            api_key (str): The API key for authentication.
            api_base (str): The base URL of the LiteLLM proxy server.
            model_name (str): The name of the model to use for generating embeddings.
            input (List[str]): A list of input strings for which embeddings are to be generated.
            timeout (float): The timeout value for the API call, default 60 secs.
            caching (bool): Whether to cache the embeddings, default True.
            **kwargs (Any): Additional keyword arguments to be passed to the embedding function.

        Returns:
            List[List[float]]: A list of embeddings, where each embedding corresponds to an input string.
        """
        try:
            from litellm import embedding
        except ImportError:
            raise ImportError(
                "To use the built-in embedding function, you need to install pytidb[models] with: "
                "pip install pytidb[models]"
            )

        response = embedding(
            input=input,
            api_key=self.api_key,
            api_base=self.api_base,
            model=self.model_name,
            timeout=self.timeout,
            caching=self.caching,
            **kwargs,
        )
        return [result["embedding"] for result in response.data]



from pytidb.orm.types import (
    BIGINT,
    BIT,
    DECIMAL,
    DOUBLE,
    FLOAT,
    INTEGER,
    INT,
    MEDIUMINT,
    NUMERIC,
    REAL,
    SMALLINT,
    TINYINT,
    CHAR,
    ENUM,
    LONGTEXT,
    MEDIUMTEXT,
    NCHAR,
    NVARCHAR,
    SET,
    TEXT,
    TINYTEXT,
    VARCHAR,
    BINARY,
    BLOB,
    LONGBLOB,
    MEDIUMBLOB,
    TINYBLOB,
    VARBINARY,
    DATE,
    DATETIME,
    TIME,
    TIMESTAMP,
    YEAR,
    JSON,
    BOOLEAN,
    VECTOR,
)

__all__ = [
    # Numeric types
    "BIGINT",
    "BIT",
    "DECIMAL",
    "DOUBLE",
    "FLOAT",
    "INTEGER",
    "INT",
    "MEDIUMINT",
    "NUMERIC",
    "REAL",
    "SMALLINT",
    "TINYINT",
    "CHAR",
    "ENUM",
    "LONGTEXT",
    "MEDIUMTEXT",
    "NCHAR",
    "NVARCHAR",
    "SET",
    "TEXT",
    "TINYTEXT",
    "VARCHAR",
    # Binary types
    "BINARY",
    "BLOB",
    "LONGBLOB",
    "MEDIUMBLOB",
    "TINYBLOB",
    "VARBINARY",
    # Date and time types
    "DATE",
    "DATETIME",
    "TIME",
    "TIMESTAMP",
    "YEAR",
    # JSON type
    "JSON",
    # Boolean type
    "BOOLEAN",
    # TiDB-specific types
    "VECTOR",
]



"""
Utility functions for embedding processing, including base64 conversion and image handling.
"""

import base64
import io
from pathlib import Path
from typing import Optional, TYPE_CHECKING, Union
from urllib.parse import ParseResult, urlparse

if TYPE_CHECKING:
    from PIL.Image import Image


def parse_url_safely(url_text: str) -> tuple[bool, Optional[ParseResult]]:
    """
    Parse a URL string and validate its format.

    Args:
        url_text: URL string to parse (should be a proper URL with scheme)

    Returns:
        Tuple of (is_valid, parsed_url) where is_valid is a boolean
        and parsed_url is the ParseResult object or None

    Note:
        This function expects properly formatted URLs with schemes (e.g., 'http://', 'https://', 'file://').
        For local file paths, use file_to_base64() or image_file_to_data_url() instead.
    """
    try:
        parsed = urlparse(url_text)
        # For file URLs, we don't require netloc
        if parsed.scheme == "file":
            is_valid = bool(parsed.scheme) and bool(parsed.path)
        else:
            is_valid = bool(parsed.scheme) and bool(parsed.netloc)
        return is_valid, parsed
    except Exception:
        return False, None


def compress_image_if_needed(
    image: "Image", max_base64_length: int = 100000
) -> "Image":
    """
    Compress an image if its base64 representation would be too large.

    Args:
        image: PIL Image object
        max_base64_length: Maximum allowed base64 string length

    Returns:
        Compressed PIL Image object
    """
    try:
        from PIL import Image
    except ImportError:
        raise ImportError(
            "PIL (Pillow) is required for image processing. Install it with: pip install Pillow"
        )

    # First try with original quality
    buffer = io.BytesIO()
    image.save(buffer, format=image.format or "JPEG", quality=95)
    base64_str = base64.b64encode(buffer.getvalue()).decode("utf-8")

    if len(base64_str) <= max_base64_length:
        return image

    # If too large, try reducing quality progressively
    for quality in [85, 70, 50, 30]:
        buffer = io.BytesIO()
        image.save(buffer, format="JPEG", quality=quality)
        base64_str = base64.b64encode(buffer.getvalue()).decode("utf-8")

        if len(base64_str) <= max_base64_length:
            # Return compressed image
            buffer.seek(0)
            return Image.open(buffer)

    # If still too large, resize the image
    original_size = image.size
    for scale in [0.8, 0.6, 0.4, 0.2]:
        new_size = (int(original_size[0] * scale), int(original_size[1] * scale))
        resized_image = image.resize(new_size, Image.Resampling.LANCZOS)

        buffer = io.BytesIO()
        resized_image.save(buffer, format="JPEG", quality=70)
        base64_str = base64.b64encode(buffer.getvalue()).decode("utf-8")

        if len(base64_str) <= max_base64_length:
            buffer.seek(0)
            return Image.open(buffer)

    # If even the smallest version is too large, return the smallest version anyway
    return resized_image


def encode_local_file_to_base64(
    file_path: Union[str, Path], max_base64_length: Optional[int] = None
) -> str:
    try:
        # Convert to Path object for better path handling
        path = Path(file_path) if isinstance(file_path, str) else file_path

        # Check if file exists
        if not path.exists():
            raise FileNotFoundError(f"File not found: {file_path}")

        # Check if it's a file (not a directory)
        if not path.is_file():
            raise ValueError(f"Path is not a file: {file_path}")

        # If a max length is provided and it's an image file, compress it
        if max_base64_length is not None and path.suffix.lower() in [
            ".jpg",
            ".jpeg",
            ".png",
            ".webp",
            ".bmp",
        ]:
            try:
                from PIL import Image

                image = Image.open(path)
                compressed_image = compress_image_if_needed(
                    image, max_base64_length=max_base64_length
                )
                buffer = io.BytesIO()
                compressed_image.save(buffer, format="JPEG", quality=70)
                return base64.b64encode(buffer.getvalue()).decode("utf-8")
            except ImportError:
                # Fall back to normal encoding if PIL is not available
                pass

        # Read file content and encode to base64
        with open(path, "rb") as file:
            buffer = io.BytesIO(file.read())
            return base64.b64encode(buffer.getvalue()).decode("utf-8")
    except Exception as e:
        raise ValueError(f"Error converting file to base64: {str(e)}")


def encode_pil_image_to_base64(
    image: "Image", max_base64_length: Optional[int] = None
) -> str:
    try:
        if max_base64_length is not None:
            image = compress_image_if_needed(image, max_base64_length=max_base64_length)

        buffer = io.BytesIO()
        image.save(buffer, format=image.format or "PNG")
        return base64.b64encode(buffer.getvalue()).decode("utf-8")
    except Exception as e:
        raise ValueError(f"Failed to encode PIL Image to base64: {str(e)}")



EMBED_PROVIDER_API_KEY_VARS = {
    # Embedding provider supported BYOK (Bring Your Own Key)
    "openai": "TIDB_EXP_EMBED_OPENAI_API_KEY",
    "jina_ai": "TIDB_EXP_EMBED_JINA_AI_API_KEY",
    "cohere": "TIDB_EXP_EMBED_COHERE_API_KEY",
    "huggingface": "TIDB_EXP_EMBED_HUGGINGFACE_API_KEY",
    "nvidia_nim": "TIDB_EXP_EMBED_NVIDIA_NIM_API_KEY",
    "gemini": "TIDB_EXP_EMBED_GEMINI_API_KEY",
}



from typing import Optional, Union, List
import numpy as np
import sqlalchemy
from sqlalchemy.dialects.mysql.base import ischema_names

# TiDB Vector has a limitation on the dimension length
MAX_DIM = 16000
MIN_DIM = 1

# Define the vector data type
VectorDataType = Union[np.ndarray, List[float]]


class VECTOR(sqlalchemy.types.UserDefinedType):
    """
    Represents a vector column type in TiDB.
    """

    dim: Optional[int]

    cache_ok = True

    def __init__(self, dim: Optional[int] = None):
        if dim is not None and not isinstance(dim, int):
            raise ValueError("expected dimension to be an integer or None")

        # tidb vector dimention length has limitation
        if dim is not None and (dim < MIN_DIM or dim > MAX_DIM):
            raise ValueError(f"expected dimension to be in [{MIN_DIM}, {MAX_DIM}]")

        super().__init__()
        self.dim = dim

    def get_col_spec(self, **kw):
        """
        Returns the column specification for the vector column.

        If the dimension is not specified, it returns "VECTOR".
        Otherwise, it returns "VECTOR(<dimension>)".

        :param kw: Additional keyword arguments.
        :return: The column specification string.
        """

        if self.dim is None:
            return "VECTOR"
        return f"VECTOR({self.dim})"

    def bind_processor(self, dialect):
        """Convert the vector float array to a string representation suitable for binding to a database column."""

        def process(value):
            return encode_vector(value, self.dim)

        return process

    def result_processor(self, dialect, coltype):
        """Convert the vector data from the database into vector array."""

        def process(value):
            return decode_vector(value)

        return process

    class comparator_factory(sqlalchemy.types.UserDefinedType.Comparator):
        """Returns a comparator factory that provides the distance functions."""

        def l1_distance(self, other: VectorDataType):
            formatted_other = encode_vector(other, self.type.dim)
            return sqlalchemy.func.VEC_L1_DISTANCE(self, formatted_other).label(
                "l1_distance"
            )

        def l2_distance(self, other: VectorDataType):
            formatted_other = encode_vector(other, self.type.dim)
            return sqlalchemy.func.VEC_L2_DISTANCE(self, formatted_other).label(
                "l2_distance"
            )

        def cosine_distance(self, other: VectorDataType):
            formatted_other = encode_vector(other, self.type.dim)
            return sqlalchemy.func.VEC_COSINE_DISTANCE(self, formatted_other).label(
                "cosine_distance"
            )

        def negative_inner_product(self, other: VectorDataType):
            formatted_other = encode_vector(other, self.type.dim)
            return sqlalchemy.func.VEC_NEGATIVE_INNER_PRODUCT(
                self, formatted_other
            ).label("negative_inner_product")

        def embed_l1_distance(self, query: str):
            return sqlalchemy.func.VEC_EMBED_L1_DISTANCE(self, query).label(
                "embed_l1_distance"
            )

        def embed_l2_distance(self, query: str):
            return sqlalchemy.func.VEC_EMBED_L2_DISTANCE(self, query).label(
                "embed_l2_distance"
            )

        def embed_cosine_distance(self, query: str):
            return sqlalchemy.func.VEC_EMBED_COSINE_DISTANCE(self, query).label(
                "embed_cosine_distance"
            )

        def embed_negative_inner_product(self, query: str):
            return sqlalchemy.func.VEC_EMBED_NEGATIVE_INNER_PRODUCT(self, query).label(
                "embed_negative_inner_product"
            )


# For reflection, make mysql dialect aware of VECTOR type.
ischema_names["vector"] = VECTOR


def encode_vector(value: VectorDataType, dim=None):
    if value is None:
        return value

    if dim is not None and len(value) != dim:
        raise ValueError(f"expected {dim} dimensions, but got {len(value)}")

    if isinstance(value, np.ndarray):
        if value.ndim != 1:
            raise ValueError("expected ndim to be 1")
        return f"[{','.join(map(str, value))}]"

    return str(value)


def decode_vector(value: str) -> np.ndarray:
    if value is None:
        return value

    if value == "[]":
        return np.array([], dtype=np.float32)

    return np.array(value[1:-1].split(","), dtype=np.float32)



import enum
from typing import Union


class DistanceMetric(str, enum.Enum):
    """
    An enumeration representing different types of distance metrics.

    - `DistanceMetric.L1`: L1 (Manhattan) distance metric.
    - `DistanceMetric.L2`: L2 (Euclidean) distance metric.
    - `DistanceMetric.COSINE`: Cosine distance metric.
    - `DistanceMetric.NEGATIVE_INNER_PRODUCT`: Negative inner product distance metric.
    """

    L1 = "L1"
    L2 = "L2"
    COSINE = "COSINE"
    NEGATIVE_INNER_PRODUCT = "NEGATIVE_INNER_PRODUCT"

    def to_sql_func(self):
        """
        Converts the DistanceMetric to its corresponding SQL function name.

        Returns:
            str: The SQL function name.

        Raises:
            ValueError: If the DistanceMetric enum member is not supported.
        """
        if self == DistanceMetric.L1:
            return "VEC_L1_DISTANCE"
        elif self == DistanceMetric.L2:
            return "VEC_L2_DISTANCE"
        elif self == DistanceMetric.COSINE:
            return "VEC_COSINE_DISTANCE"
        elif self == DistanceMetric.NEGATIVE_INNER_PRODUCT:
            return "VEC_NEGATIVE_INNER_PRODUCT"
        else:
            raise ValueError(
                f"Distance metric {self} has no corresponding SQL function"
            )


def validate_distance_metric(value: Union[str, DistanceMetric]) -> DistanceMetric:
    """
    Validate and convert distance metric parameter.

    Args:
        value: String or DistanceMetric enum to validate and convert

    Returns:
        DistanceMetric: The validated and converted enum value

    Raises:
        ValueError: If the value is not a valid distance metric

    Examples:
        >>> validate_distance_metric("COSINE")
        DistanceMetric.COSINE
        >>> validate_distance_metric("l2")
        DistanceMetric.L2
        >>> validate_distance_metric(DistanceMetric.COSINE)
        DistanceMetric.COSINE
    """
    if isinstance(value, DistanceMetric):
        return value
    elif isinstance(value, str):
        value_upper = value.upper()
        if value_upper == "L1":
            return DistanceMetric.L1
        elif value_upper == "L2":
            return DistanceMetric.L2
        elif value_upper == "COSINE":
            return DistanceMetric.COSINE
        elif value_upper == "NEGATIVE_INNER_PRODUCT":
            return DistanceMetric.NEGATIVE_INNER_PRODUCT
        else:
            raise ValueError(
                f"Invalid distance metric: {value}. Valid options: L1, L2, COSINE, NEGATIVE_INNER_PRODUCT"
            )
    else:
        raise ValueError(
            f"Invalid distance metric type: {type(value)}. Expected string or DistanceMetric enum"
        )



import os

from sqlmodel import Session
from sqlalchemy import create_engine

from .client import TiDBClient
from .table import Table
from .utils import build_tidb_connection_url

if "LITELLM_LOCAL_MODEL_COST_MAP" not in os.environ:
    os.environ["LITELLM_LOCAL_MODEL_COST_MAP"] = "True"

if "LITELLM_LOG" not in os.environ:
    os.environ["LITELLM_LOG"] = "WARNING"


__all__ = [
    "TiDBClient",
    "Table",
    "build_tidb_connection_url",
    "Session",
    "create_engine",
]



import re

from urllib.parse import quote
from typing import Dict, Optional, Any, List, TypeVar, Tuple

from pydantic import AnyUrl, UrlConstraints
from sqlalchemy import Column, Index, String, create_engine, make_url
from sqlmodel import AutoString
from sqlalchemy.engine import Row
from sqlalchemy import Table
from typing import Union
from pytidb.orm.vector import VECTOR

TIDB_SERVERLESS_HOST_PATTERN = re.compile(
    r"gateway\d{2}\.(.+)\.(prod|dev|staging)\.(shared\.)?(aws|alicloud)\.tidbcloud\.com"
)


def create_engine_without_db(url, echo=False, **kwargs):
    temp_db_url = make_url(url)
    temp_db_url = temp_db_url._replace(database=None)
    return create_engine(temp_db_url, echo=echo, **kwargs)


class TiDBConnectionURL(AnyUrl):
    """A URL that enforces specific constraints for TiDB connections.

    * User info required
    * TLD not required
    * Host not required
    """

    _constraints = UrlConstraints(
        allowed_schemes=[
            "mysql",
            "mysql+mysqlconnector",
            "mysql+aiomysql",
            "mysql+asyncmy",
            "mysql+mysqldb",
            "mysql+pymysql",
            "mysql+cymysql",
            "mysql+pyodbc",
        ],
        default_port=4000,
        host_required=True,
    )


def build_tidb_connection_url(
    schema: str = "mysql+pymysql",
    host: str = "localhost",
    port: int = 4000,
    username: str = "root",
    password: str = "",
    database: str = "test",
    enable_ssl: Optional[bool] = None,
) -> str:
    """
    Build a TiDB Connection URL string for database connection.

    Args:
        schema (str, optional): The connection protocol. Defaults to "mysql+pymysql".
        host (str, optional): The host address of TiDB server. Defaults to "localhost".
        port (int, optional): The port number of TiDB server. Defaults to 4000.
        username (str, optional): The username for authentication. Defaults to "root".
        password (str, optional): The password for authentication. Defaults to "".
        database (str, optional): The database name to connect to. Defaults to "test".
        enable_ssl (Optional[bool], optional): Whether to enable SSL for the connection.
            If None (default), SSL is automatically enabled for TiDB Serverless hosts
            and disabled for other hosts.

    Returns:
        str: A Connection URL string that can be used to connect to a TiDB database.
    """

    if enable_ssl is None:
        if host and TIDB_SERVERLESS_HOST_PATTERN.match(host):
            enable_ssl = True
        else:
            enable_ssl = None

    return str(
        TiDBConnectionURL.build(
            scheme=schema,
            host=host,
            port=port,
            username=username,
            # TODO: remove quote after following issue is fixed:
            # https://github.com/pydantic/pydantic/issues/8061
            password=quote(password) if password else None,
            path=database,
            query=(
                "ssl_verify_cert=true&ssl_verify_identity=true" if enable_ssl else None
            ),
        )
    )


def filter_vector_columns(columns: Dict) -> List[Column]:
    vector_columns = []
    for column in columns:
        if isinstance(column.type, VECTOR):
            vector_columns.append(column)
    return vector_columns


def check_vector_column(columns: Dict, column_name: str) -> Optional[Column]:
    if not isinstance(column_name, str):
        raise ValueError(f"Invalid vector column name: {column_name}")

    if column_name not in columns:
        raise ValueError(f"Non-exists vector column: {column_name}")

    vector_column = columns[column_name]
    if not isinstance(vector_column.type, VECTOR):
        raise ValueError(f"Invalid vector column: {column_name}")

    return vector_column


def filter_text_columns(columns: Dict) -> List[Column]:
    text_columns = []
    for column in columns:
        if isinstance(column.type, AutoString) or isinstance(column.type, String):
            text_columns.append(column)
    return text_columns


def check_text_column(columns: Dict, column_name: str) -> Optional[str]:
    if column_name not in columns:
        raise ValueError(f"Non-exists text column: {column_name}")

    text_column = columns[column_name]
    if not isinstance(text_column.type, String) and not isinstance(
        text_column.type, AutoString
    ):
        raise ValueError(f"Invalid text column: {text_column}")

    return text_column


RowKeyType = TypeVar("RowKeyType", bound=Union[Any, Tuple[Any, ...]])


def get_row_id_from_row(row: Row, table: Table) -> Optional[RowKeyType]:
    pk_constraint = table.primary_key
    if not pk_constraint.columns:
        # Try to get _tidb_rowid if no primary key exists
        try:
            return row._mapping["_tidb_rowid"]
        except KeyError:
            return row.__hash__()

    pk_column_names = [col.name for col in pk_constraint.columns]
    try:
        if len(pk_column_names) == 1:
            return row._mapping[pk_column_names[0]]
        return tuple(row._mapping[name] for name in pk_column_names)
    except KeyError as e:
        raise KeyError(
            f"Primary key column '{e.args[0]}' not found in Row. "
            f"Available: {list(row._mapping.keys())}"
        )


def get_index_type(index: Index) -> str:
    dialect_kwargs = getattr(index, "dialect_kwargs", None)
    if dialect_kwargs is None:
        return ""
    mysql_prefix = dialect_kwargs.get("mysql_prefix", "")
    return mysql_prefix.lower() if mysql_prefix else ""
